@@ -73,47 +73,14 @@ InputCapturePortal::InputCapturePortal(QObject *parent)
     qDBusRegisterMetaType<QList<QPair<QPoint, QPoint>>>();
 }
 
-uint InputCapturePortal::CreateSession(const QDBusObjectPath &handle,
-                                       const QDBusObjectPath &session_handle,
-                                       const QString &app_id,
-                                       const QString &parent_window,
-                                       const QVariantMap &options,
-                                       QVariantMap &results)
+bool InputCapturePortal::setupInputCaptureSession(InputCaptureSession *session, Capabilities capabilities)
 {
-    qCDebug(XdgDesktopPortalKdeInputCapture) << "CreateSession called with parameters:";
-    qCDebug(XdgDesktopPortalKdeInputCapture) << "    handle: " << handle.path();
-    qCDebug(XdgDesktopPortalKdeInputCapture) << "    session_handle: " << session_handle.path();
-    qCDebug(XdgDesktopPortalKdeInputCapture) << "    app_id: " << app_id;
-    qCDebug(XdgDesktopPortalKdeInputCapture) << "    parent_window: " << parent_window;
-    qCDebug(XdgDesktopPortalKdeInputCapture) << "    options: " << options;
-
-    auto *session = static_cast<InputCaptureSession *>(Session::createSession(this, Session::InputCapture, app_id, session_handle.path()));
-
-    if (!session) {
-        return 2;
-    }
-
-    const auto requestedCapabilities = options.value("capabilities").toUInt();
-    if (requestedCapabilities == 0) {
-        qCWarning(XdgDesktopPortalKdeInputCapture) << "No capabilities requested";
-        return 2;
-    }
-
-    InputCaptureDialog dialog(app_id, Capabilities::fromInt(requestedCapabilities), this);
-    Utils::setParentWindow(dialog.windowHandle(), parent_window);
-    Request::makeClosableDialogRequest(handle, &dialog);
-    connect(session, &Session::closed, &dialog, &QuickDialog::reject);
-
-    if (!dialog.exec()) {
-        return 1;
-    }
-
     auto msg = QDBusMessage::createMethodCall(kwinService(), kwinInputCapturePath(), kwinInputCaptureManagerInterface(), u"addInputCapture"_s);
-    msg << static_cast<int>(requestedCapabilities);
+    msg << capabilities.toInt();
     QDBusReply<QDBusObjectPath> reply = QDBusConnection::sessionBus().call(msg, QDBus::Block, kwinDBusTimeout);
     if (!reply.isValid()) {
         qCWarning(XdgDesktopPortalKdeInputCapture) << "Failed to create KWin input capture:" << reply.error();
-        return 2;
+        return false;
     }
     session->connect(reply.value());
 
@@ -154,15 +121,61 @@ uint InputCapturePortal::CreateSession(const QDBusObjectPath &handle,
                                          "Input is being managed by an application. Press <shortcut>%1</shortcut> to disable.",
                                          disableSequence.toString(QKeySequence::NativeText)));
         }
-        notification->setIconName("dialog-input-devices");
+        notification->setIconName(u"dialog-input-devices"_s);
         connect(session, &InputCaptureSession::deactivated, notification, &KNotification::close);
         notification->sendEvent();
 
         Q_EMIT Activated(QDBusObjectPath(session->handle()), {{u"activation_id"_s, activationId}, {u"cursor_position"_s, cursorPosition}});
     });
+    return true;
+}
 
-    results.insert(u"capabilities"_s, requestedCapabilities);
-    return 0;
+void InputCapturePortal::CreateSession(const QDBusObjectPath &handle,
+                                       const QDBusObjectPath &session_handle,
+                                       const QString &app_id,
+                                       const QString &parent_window,
+                                       const QVariantMap &options,
+                                       const QDBusMessage &message,
+                                       uint &replyResponse,
+                                       [[maybe_unused]] QVariantMap &replyResults)
+{
+    qCDebug(XdgDesktopPortalKdeInputCapture) << "CreateSession called with parameters:";
+    qCDebug(XdgDesktopPortalKdeInputCapture) << "    handle: " << handle.path();
+    qCDebug(XdgDesktopPortalKdeInputCapture) << "    session_handle: " << session_handle.path();
+    qCDebug(XdgDesktopPortalKdeInputCapture) << "    app_id: " << app_id;
+    qCDebug(XdgDesktopPortalKdeInputCapture) << "    parent_window: " << parent_window;
+    qCDebug(XdgDesktopPortalKdeInputCapture) << "    options: " << options;
+
+    auto *session = static_cast<InputCaptureSession *>(Session::createSession(this, Session::InputCapture, app_id, session_handle.path()));
+
+    if (!session) {
+        replyResponse = PortalResponse::OtherError;
+        return;
+    }
+
+    const auto requestedCapabilities = Capabilities::fromInt(options.value(u"capabilities"_s).toUInt());
+    if (requestedCapabilities == Capability::None) {
+        qCWarning(XdgDesktopPortalKdeInputCapture) << "No capabilities requested";
+        replyResponse = PortalResponse::OtherError;
+        return;
+    }
+
+    auto dialog = new InputCaptureDialog(app_id, (requestedCapabilities), this);
+    Utils::setParentWindow(dialog->windowHandle(), parent_window);
+    Request::makeClosableDialogRequestWithSession(handle, dialog, session);
+
+    delayReply(message, dialog, this, [this, session, requestedCapabilities](DialogResult result) {
+        auto response = PortalResponse::fromDialogResult(result);
+        QVariantMap results;
+        if (result == DialogResult::Accepted) {
+            if (!setupInputCaptureSession(session, requestedCapabilities)) {
+                response = PortalResponse::OtherError;
+            } else {
+                results.insert(u"capabilities"_s, static_cast<uint>(requestedCapabilities));
+            }
+        }
+        return QVariantList{response, results};
+    });
 }
 
 uint InputCapturePortal::GetZones(const QDBusObjectPath &handle,
@@ -178,11 +191,11 @@ uint InputCapturePortal::GetZones(const QDBusObjectPath &handle,
     qCDebug(XdgDesktopPortalKdeInputCapture) << "    app_id: " << app_id;
     qCDebug(XdgDesktopPortalKdeInputCapture) << "    options: " << options;
 
-    auto *session = qobject_cast<InputCaptureSession *>(Session::getSession(session_handle.path()));
+    auto *session = Session::getSession<InputCaptureSession>(session_handle.path());
 
     if (!session) {
         qCWarning(XdgDesktopPortalKdeInputCapture) << "Tried to get zones on non-existing session " << session_handle.path();
-        return 2;
+        return PortalResponse::OtherError;
     }
 
     auto handleZoneChange = [this, session] {
@@ -199,7 +212,7 @@ uint InputCapturePortal::GetZones(const QDBusObjectPath &handle,
         Q_EMIT ZonesChanged(QDBusObjectPath(session->handle()), {{u"zone_set"_s, m_zoneId}});
     };
 
-    results.insert("zone_set", m_zoneId);
+    results.insert(u"zone_set"_s, m_zoneId);
     QList<zone> zones;
     for (const auto screen : qGuiApp->screens()) {
         zones.push_back(zone{
@@ -214,8 +227,8 @@ uint InputCapturePortal::GetZones(const QDBusObjectPath &handle,
     connect(qGuiApp, &QGuiApplication::screenAdded, session, handleZoneChange);
     connect(qGuiApp, &QGuiApplication::screenRemoved, session, handleZoneChange);
 
-    results.insert("zones", QVariant::fromValue(zones));
-    return 0;
+    results.insert(u"zones"_s, QVariant::fromValue(zones));
+    return PortalResponse::Success;
 }
 
 uint InputCapturePortal::SetPointerBarriers(const QDBusObjectPath &handle,
@@ -234,22 +247,22 @@ uint InputCapturePortal::SetPointerBarriers(const QDBusObjectPath &handle,
     qCDebug(XdgDesktopPortalKdeInputCapture) << "    zone_set: " << zone_set;
     qCDebug(XdgDesktopPortalKdeInputCapture) << "    barriers: ";
 
-    auto *session = qobject_cast<InputCaptureSession *>(Session::getSession(session_handle.path()));
+    auto *session = Session::getSession<InputCaptureSession>(session_handle.path());
 
     if (!session) {
         qCWarning(XdgDesktopPortalKdeInputCapture) << "Tried to set barriers non-existing session " << session_handle.path();
-        return 2;
+        return PortalResponse::OtherError;
     }
 
     if (zone_set != m_zoneId) {
         qCWarning(XdgDesktopPortalKdeInputCapture) << "Invalid zone_set " << session_handle.path();
-        return 2;
+        return PortalResponse::OtherError;
     }
 
     if (session->state != State::Disabled) {
         if (auto reply = QDBusReply(session->disable()); !reply.isValid()) {
             qCWarning(XdgDesktopPortalKdeInputCapture) << "Error disabling input capture:" << reply.error();
-            return 2;
+            return PortalResponse::OtherError;
         }
     }
     session->clearBarriers();
@@ -297,7 +310,7 @@ uint InputCapturePortal::SetPointerBarriers(const QDBusObjectPath &handle,
         }
     }
     results.insert(u"failed_barriers"_s, QVariant::fromValue(failedBarriers));
-    return 0;
+    return PortalResponse::Success;
 }
 
 QDBusUnixFileDescriptor
@@ -308,7 +321,7 @@ InputCapturePortal::ConnectToEIS(const QDBusObjectPath &session_handle, const QS
     qCDebug(XdgDesktopPortalKdeInputCapture) << "    app_id: " << app_id;
     qCDebug(XdgDesktopPortalKdeInputCapture) << "    options: " << options;
 
-    auto *session = qobject_cast<InputCaptureSession *>(Session::getSession(session_handle.path()));
+    auto *session = Session::getSession<InputCaptureSession>(session_handle.path());
     if (!session) {
         qCWarning(XdgDesktopPortalKdeInputCapture) << "Tried to call ConnectToEis on non-existing session " << session_handle.path();
         return QDBusUnixFileDescriptor();
@@ -340,25 +353,25 @@ uint InputCapturePortal::Enable(const QDBusObjectPath &session_handle, const QSt
     qCDebug(XdgDesktopPortalKdeInputCapture) << "    app_id: " << app_id;
     qCDebug(XdgDesktopPortalKdeInputCapture) << "    options: " << options;
 
-    auto *session = qobject_cast<InputCaptureSession *>(Session::getSession(session_handle.path()));
+    auto *session = Session::getSession<InputCaptureSession>(session_handle.path());
     if (!session) {
         qCWarning(XdgDesktopPortalKdeInputCapture) << "Tried to call Enable on non-existing session " << session_handle.path();
-        return 2;
+        return PortalResponse::OtherError;
     }
 
     if (session->state != State::Disabled) {
         qCWarning(XdgDesktopPortalKdeInputCapture) << "Session is already enabled" << session_handle.path();
-        return 2;
+        return PortalResponse::OtherError;
     }
 
     QDBusReply reply = session->enable();
     if (!reply.isValid()) {
         qCWarning(XdgDesktopPortalKdeInputCapture) << "Failed to enable session" << reply.error();
-        return 2;
+        return PortalResponse::OtherError;
     }
 
     session->state = State::Deactivated;
-    return 0;
+    return PortalResponse::Success;
 }
 
 uint InputCapturePortal::Disable(const QDBusObjectPath &session_handle, const QString &app_id, const QVariantMap &options, QVariantMap &results)
@@ -369,24 +382,24 @@ uint InputCapturePortal::Disable(const QDBusObjectPath &session_handle, const QS
     qCDebug(XdgDesktopPortalKdeInputCapture) << "    app_id: " << app_id;
     qCDebug(XdgDesktopPortalKdeInputCapture) << "    options: " << options;
 
-    auto *session = qobject_cast<InputCaptureSession *>(Session::getSession(session_handle.path()));
+    auto *session = Session::getSession<InputCaptureSession>(session_handle.path());
     if (!session) {
         qCWarning(XdgDesktopPortalKdeInputCapture) << "Tried to call Enable on non-existing session " << session_handle.path();
-        return 2;
+        return PortalResponse::OtherError;
     }
 
     if (session->state == State::Disabled) {
         qCWarning(XdgDesktopPortalKdeInputCapture) << "Session is not enabled" << session_handle.path();
-        return 2;
+        return PortalResponse::OtherError;
     }
 
     QDBusReply reply = session->enable();
     if (!reply.isValid()) {
         qCWarning(XdgDesktopPortalKdeInputCapture) << "Failed to disable session" << reply.error();
-        return 2;
+        return PortalResponse::OtherError;
     }
 
-    return 0;
+    return PortalResponse::Success;
 }
 
 uint InputCapturePortal::Release(const QDBusObjectPath &session_handle, const QString &app_id, const QVariantMap &options, QVariantMap &results)
@@ -397,18 +410,18 @@ uint InputCapturePortal::Release(const QDBusObjectPath &session_handle, const QS
     qCDebug(XdgDesktopPortalKdeInputCapture) << "    app_id: " << app_id;
     qCDebug(XdgDesktopPortalKdeInputCapture) << "    options: " << options;
 
-    auto *session = qobject_cast<InputCaptureSession *>(Session::getSession(session_handle.path()));
+    auto *session = Session::getSession<InputCaptureSession>(session_handle.path());
     if (!session) {
         qCWarning(XdgDesktopPortalKdeInputCapture) << "Tried to call Enable on non-existing session " << session_handle.path();
-        return 2;
+        return PortalResponse::OtherError;
     }
 
     if (session->state != State::Activated) {
         qCWarning(XdgDesktopPortalKdeInputCapture) << "Session is not activated" << session_handle.path();
-        return 2;
+        return PortalResponse::OtherError;
     }
 
-    auto it = options.find("cursor_position");
+    auto it = options.find(u"cursor_position"_s);
     bool positionSpecified = it != options.end();
     QPointF cursorPosition = positionSpecified ? qdbus_cast<QPointF>(it->value<QDBusArgument>()) : QPointF(); // (dd)
 
@@ -421,8 +434,8 @@ uint InputCapturePortal::Release(const QDBusObjectPath &session_handle, const QS
     QDBusReply reply = session->release(cursorPosition, positionSpecified);
     if (!reply.isValid()) {
         qCWarning(XdgDesktopPortalKdeInputCapture) << "Failed to release session" << reply.error();
-        return 2;
+        return PortalResponse::OtherError;
     }
 
-    return 0;
+    return PortalResponse::Success;
 }
