@@ -20,17 +20,15 @@ class NotificationPortal : public QDBusAbstractAdaptor
     Q_CLASSINFO("D-Bus Interface", "org.freedesktop.impl.portal.Notification")
 public:
     explicit NotificationPortal(QObject *parent);
-    ~NotificationPortal() override;
 
 public Q_SLOTS:
     void AddNotification(const QString &app_id, const QString &id, const QVariantMap &notification);
     void RemoveNotification(const QString &app_id, const QString &id);
 private Q_SLOTS:
-    void notificationActivated(uint action);
     void notificationClosed();
 
 private:
-    QHash<QString, KNotification *> m_notifications;
+    QHash<QString, QPointer<KNotification>> m_notifications;
 };
 
 #endif // XDG_DESKTOP_PORTAL_KDE_NOTIFICATION_H
